/* usadelndsoc
 *
 * Copyright © 2022 Pauli Virtanen
 *    @author Pauli Virtanen <pauli.t.virtanen@jyu.fi>
 *
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */
#ifndef ACTION_HPP_
#define ACTION_HPP_

#include <Eigen/Core>
#include <Eigen/LU>

#include "common.hpp"
#include "array.hpp"

template <typename M> inline auto block_00(M& mat) { return mat.template block<2,2>(0,0); }
template <typename M> inline auto block_01(M& mat) { return mat.template block<2,2>(0,2); }
template <typename M> inline auto block_10(M& mat) { return mat.template block<2,2>(2,0); }
template <typename M> inline auto block_11(M& mat) { return mat.template block<2,2>(2,2); }

template <typename ConstScalar, typename Scalar = typename std::remove_const<ConstScalar>::type>
inline Matrix<Scalar,4,4>
get_Q_matrix(const array::ArrayView<ConstScalar, array::Shape<2,2,2> >& Q)
{
    auto g = Q.part(0u).matrix();
    auto gt = Q.part(1u).matrix();
    Matrix<Complex,2,2> I;
    Matrix<Scalar,2,2> N, Nt;
    Matrix<Scalar,4,4> Qm;

    I << 1, 0,
        0, 1;

    N = (I + g*gt).inverse();
    Nt = (I + gt*g).inverse();

    block_00(Qm) = N * (I - g * gt);
    block_01(Qm) = 2 * N * g;
    block_10(Qm) = 2 * Nt * gt;
    block_11(Qm) = -Nt * (I - gt * g);

    return Qm;
}

template <typename LRType, typename XType,
          typename ResultScalar=decltype(typename XType::value_type()*typename LRType::value_type())>
Eigen::Matrix<ResultScalar,4,4> nambu_diag_mul(const LRType& L, const XType& X, const LRType& R)
{
    /* Multiply L*X*R with Nambu-diagonal L, R */
    Eigen::Matrix<ResultScalar,4,4> Y;
    block_00(Y) = block_00(L) * block_00(X) * block_00(R);
    block_01(Y) = block_00(L) * block_01(X) * block_11(R);
    block_10(Y) = block_11(L) * block_10(X) * block_00(R);
    block_11(Y) = block_11(L) * block_11(X) * block_11(R);
    return Y;
}

template <typename Scalar,
          typename ConstScalarQ, typename ConstScalarU,
          typename Shape0, typename Shape, typename Shape2>
inline Scalar S_2(const array::ArrayView<const Mask, Shape0>& mask,
                  const array::ArrayView<ConstScalarQ, Shape>& Qmat,
                  const array::ArrayView<ConstScalarU, Shape2>& U,
                  const double& Lx,
                  const double& Ly,
                  const Complex& D)
{
    const int nx = Qmat.dim(0);
    const int ny = Qmat.dim(1);

    const double dd[2] = {Lx / nx, Ly / ny};

    const int N_i[2] = {1, 0};
    const int N_j[2] = {0, 1};

    /* Directions cycle (left->right, down->up, right->left, up->down) */
    const size_t invdir[4] = {2, 3, 0, 1};
    Scalar S{0.0};

    if (D == Complex(0))
        return S;

    for (int i = 0; i < nx; ++i) {
        for (int j = 0; j < ny; ++j) {
            if (mask(i,j) == Mask::VACUUM)
                continue;

            auto Q1 = Qmat.part(i,j).matrix();

            /* Sum over neighbors */
            for (size_t k = 0; k < 2; ++k) {
                int i2 = i + N_i[k];
                int j2 = j + N_j[k];
                
                if (i2 < 0 || i2 >= nx || j2 < 0 || j2 >= ny)
                    continue;
                if (mask(i2,j2) == Mask::VACUUM)
                    continue;

                auto Q2 = Qmat.part(i2,j2).matrix();
                auto U12 = U.part(i,j,k).matrix();
                auto U21 = U.part(i2,j2,invdir[k]).matrix();

                auto M = nambu_diag_mul(U12, Q2, U21) - Q1;

                S += (M * M).trace() * (1 / (dd[k] * dd[k]));
            }
        }
    }

    return (D*dd[0]*dd[1]*2.0/2.0) * S;
}

template <typename Scalar,
          typename ConstScalarQ, typename ConstScalarOmega,
          typename Shape0, typename Shape, typename Shape2>
inline Scalar S_Omega(const array::ArrayView<const Mask, Shape0>& mask,
                      const array::ArrayView<ConstScalarQ, Shape>& Qmat,
                      const Scalar& omega,
                      const array::ArrayView<ConstScalarOmega, Shape2>& Omega,
                      const double& Lx,
                      const double& Ly)
{
    Scalar S{0.0};

    const int nx = Qmat.dim(0);
    const int ny = Qmat.dim(1);

    const double dd[2] = {Lx / nx, Ly / ny};

    for (int i = 0; i < nx; ++i) {
        for (int j = 0; j < ny; ++j) {
            if (mask(i,j) == Mask::VACUUM)
                continue;

            auto Q = Qmat.part(i,j).matrix();
            auto w = Omega.part(i, j).matrix();

            S += (w * Q).trace();
            S += (Complex(0, 1) * omega) * (Q(0,0) + Q(1,1) - Q(2,2) - Q(3,3));
        }
    }

    return (Complex(0, 4) * dd[0]*dd[1]) * S;
}

template <typename Scalar, typename ConstScalarQ, typename ConstScalarU, typename Shape0, typename ShapeQ, typename ShapeU>
inline Scalar S_H(const array::ArrayView<const Mask, Shape0>& mask,
                  const array::ArrayView<ConstScalarQ, ShapeQ>& Qmat,
                  const array::ArrayView<ConstScalarU, ShapeU>& U,
                  const double& Lx,
                  const double& Ly,
                  const Complex& alpha)
{
    using array::Array;
    using array::Shape;

    const int nx = Qmat.dim(0);
    const int ny = Qmat.dim(1);

    const double dd[2] = {Lx / nx, Ly / ny};

    /* Plaquettes: (dir) -> (di,dj)
     * where 'dir' is the direction index of the link,
     * starting from bottom and cycling counterclockwise.
     */
    const int plaq[4][2] = {{0,0}, {1,0}, {1,1}, {0,1}};

    const size_t invdir[4] = {2, 3, 0, 1};

    Scalar S{0.0};

    if (alpha == Complex(0))
        return S;

    for (int i = 0; i < nx-1; ++i) {
        for (int j = 0; j < ny-1; ++j) {
            if (mask(i,j) == Mask::VACUUM)
                continue;

            /* Plaquette with lower left corner at (i,j) */

            /* Sum over cyclic permutations */
            for (int iperm = 0; iperm < 4; ++iperm) {
                const int k1 = (iperm+0)%4;
                const int k2 = (iperm+1)%4;
                const int k3 = (iperm+2)%4;
                const int k4 = (iperm+3)%4;

                const int i1 = (int)i + plaq[k1][0];
                const int j1 = (int)j + plaq[k1][1];
                const int i2 = (int)i + plaq[k2][0];
                const int j2 = (int)j + plaq[k2][1];
                const int i3 = (int)i + plaq[k3][0];
                const int j3 = (int)j + plaq[k3][1]; 
                const int i4 = (int)i + plaq[k4][0];
                const int j4 = (int)j + plaq[k4][1];

                if (mask(i1,j1) == Mask::VACUUM)
                    continue;
                if (mask(i2,j2) == Mask::VACUUM)
                    continue;
                if (mask(i3,j3) == Mask::VACUUM)
                    continue;
                if (mask(i4,j4) == Mask::VACUUM)
                    continue;

                auto Q1 = Qmat.part(i1,j1).matrix();
                auto Q2 = Qmat.part(i2,j2).matrix();
                //auto Q3 = Qmat.part(i3,j3).matrix();
                auto Q4 = Qmat.part(i4,j4).matrix();

                auto U21 = U.part(i1,j1,k1).matrix();
                auto U32 = U.part(i2,j2,k2).matrix();
                auto U43 = U.part(i3,j3,k3).matrix();
                auto U14 = U.part(i4,j4,k4).matrix();

                auto U12 = U.part(i2,j2,invdir[k1]).matrix();
                //auto U23 = U.part(i3,j3,invdir[k2]).matrix();
                //auto U34 = U.part(i4,j4,invdir[k3]).matrix();
                auto U41 = U.part(i1,j1,invdir[k4]).matrix();

                auto P = U14*U43*U32*U21;

                Matrix<Complex,4,4> I;

                I << 1, 0, 0, 0,
                    0, 1, 0, 0,
                    0, 0, 1, 0,
                    0, 0, 0, 1;

                S += ((P - I) * (Q1 - nambu_diag_mul(U12, Q2, U21) * Q1 * nambu_diag_mul(U14, Q4, U41))).trace();
            }
        }
    }

    return (2.0*alpha/(dd[0]*dd[1])) * S;
}

/* Function that has the same value as the AD variable,
   but has zero derivatives */
Complex flat_value(const Complex& value)
{
    return value;
}
CPPAD_DISCRETE_FUNCTION(Complex, flat_value);

template <typename Scalar,
    typename ConstScalarQ, typename ConstScalarU, typename ConstScalarOmega,
    typename Shape0, typename ShapeQ, typename ShapeU, typename ShapeOmega,
    typename ScalarQ = typename std::remove_const<ConstScalarQ>::type>
inline Scalar S(const array::ArrayView<const Mask, Shape0>& mask,
                const array::ArrayView<ConstScalarQ, ShapeQ>& Q,
                const Scalar& omega,
                const array::ArrayView<ConstScalarU, ShapeU>& U,
                const array::ArrayView<ConstScalarOmega, ShapeOmega>& Omega,
                const double& Lx,
                const double& Ly,
                const Complex& alpha,
                const Complex& D)
{
    using array::Array;
    using array::Shape;
    using array::Dynamic;

    const size_t nx = Q.dim(0);
    const size_t ny = Q.dim(1);

    Scalar S{0.0};

    Array<ScalarQ, Shape<Dynamic, Dynamic, 4, 4> > Qmat({nx, ny, 4, 4});

    for (size_t i = 0; i < nx; ++i) {
        for (size_t j = 0; j < ny; ++j) {
            if (mask(i,j) == Mask::VACUUM) {
                continue;
            } else if (mask(i,j) == Mask::TERMINAL) {
                const ScalarQ *Qp = Q.part(i,j).data();
                Array<ScalarQ,Shape<2,2,2> > Qpart;

                for (size_t p = 0; p < Qpart.size(); ++p) {
                    auto w = flat_value(Qp[p]);

                    Qpart.data()[p] = w;

                    /* Pin value to the flat value */
                    auto u = Qp[p] - w;
                    S = S + 0.5*u*u;
                }

                Qmat.part(i,j).matrix() = get_Q_matrix(Qpart);
            } else {
                Qmat.part(i,j).matrix() = get_Q_matrix(Q.part(i,j));
            }
        }
    }

    S = S + S_2<Scalar>(mask, Qmat, U, Lx, Ly, D) + S_Omega<Scalar>(mask, Qmat, omega, Omega, Lx, Ly) + S_H<Scalar>(mask, Qmat, U, Lx, Ly, alpha);
    return S;
}

#endif

